import os.path
import unittest
from natural_selection.genetic_algorithms import Gene, Chromosome, Individual, Island
from natural_selection.genetic_algorithms.utils.random_functions import random_int
from natural_selection.genetic_algorithms.operators.crossover import crossover_two_n_point

class TestSimpleIsland(unittest.TestCase):

    def setUp(self) -> None:
        self.fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])
        self.ind = Individual(self.fitness, name="Adam", chromosome=gen)
        self.ind.add_new_property('some_property', 10)

        self.life = Island({'x': 0.6, 'y' : 0.2}, crossover_function=crossover_two_n_point)

    def test_write_report(self):
        self.life.initialise(self.ind, population_size=5)
        self.life.evolve(crossover_params={'n_points': 2})

        self.life.write_report('test.csv')
        self.life.write_report('test.json', output_json=True)

        self.assertTrue(os.path.isfile('test.csv'))
        self.assertTrue(os.path.isfile('test.json'))

        os.remove('test.csv')
        os.remove('test.json')

    def test_lineage(self):
        self.life.initialise(self.ind, population_size=5)
        self.life.evolve(crossover_params={'n_points': 2}, n_generations=10, crossover_probability=0.6)

        self.assertTrue(len(self.life.lineage['edges']) > 1)

        self.life.write_lineage_json('test.json')

        self.assertTrue(os.path.isfile('test.json'))

        os.remove('test.json')

    def test_create(self):
        self.life.initialise(self.ind, population_size=5)
        self.assertEquals(len(self.life.population), 5)

    def test_import_migrants(self):
        self.life.initialise(self.ind, population_size=5)

        fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])

        aliens = [Individual(fitness, name="AlsoAdam", chromosome=gen)]

        self.life.import_migrants(aliens)

        self.assertEquals(len(self.life.population), 5)

        fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=5, gene_max=10, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=99, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])

        aliens = [Individual(fitness, name="Eve", chromosome=gen)]

        self.life.import_migrants(aliens)

        self.assertEquals(len(self.life.population), 6)

    def test_evolve_generational(self):
        self.life.initialise(self.ind, population_size=5)
        self.life.evolve(crossover_params={'n_points' : 2})

class TestSimpleIslandMinimise(unittest.TestCase):

    def setUp(self) -> None:
        self.fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=1, gene_max=10, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=10, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])
        self.ind = Individual(self.fitness, name="Adam", chromosome=gen)
        self.ind.add_new_property('some_property', 10)

        self.life = Island({'x': 0.6, 'y' : 0.2}, maximise_function=False)

    def test_evolve_generational(self):
        self.life.initialise(self.ind, population_size=5)
        best = self.life.evolve()

        self.assertLessEqual(best.fitness, 8)

class TestSimpleIslandMaximise(unittest.TestCase):

    def setUp(self) -> None:
        self.fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=1, gene_max=10, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=10, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])
        self.ind = Individual(self.fitness, name="Adam", chromosome=gen)
        self.ind.add_new_property('some_property', 10)

        self.life = Island({'x': 0.6, 'y' : 0.2})

    def test_evolve_generational(self):
        self.life.initialise(self.ind, population_size=5)
        best = self.life.evolve()

        self.assertGreaterEqual(best.fitness, 0.8)

class TestOtherIsland(unittest.TestCase):

    def setUp(self) -> None:
        self.fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])
        self.ind = Individual(self.fitness, name="Adam", chromosome=gen)

        self.life = Island({'x': 0.6, 'y' : 0.2}, crossover_function=crossover_two_n_point)

    def test_create(self):
        self.life.initialise(self.ind, population_size=5)
        self.assertEquals(len(self.life.population), 5)

    def test_import_migrants(self):
        self.life.initialise(self.ind, population_size=5)

        fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])

        aliens = [Individual(fitness, name="AlsoAdam", chromosome=gen)]

        self.life.import_migrants(aliens)

        self.assertEquals(len(self.life.population), 5)

        fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)
        g_1 = Gene(name="first", value=5, gene_max=10, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=99, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])

        aliens = [Individual(fitness, name="Eve", chromosome=gen)]

        self.life.import_migrants(aliens)

        self.assertEquals(len(self.life.population), 6)

    def test_evolve_generational(self):
        self.life.initialise(self.ind, population_size=5)
        self.life.evolve()

def fitness(individual, island, x, y):
    return (individual.chromosome[0].value * x) + (individual.chromosome[1].value * y)

def fitness_raise_error(individual, island, x, y):
    return 1 / 0

class TestSaveLoad(unittest.TestCase):

    def test_save_load(self):
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])
        ind = Individual(fitness, name="Adam", chromosome=gen)

        life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point)

        life.initialise(ind, population_size=5)

        life.evolve()

        fp = 'testing_island.pkl'

        life.save_island(fp)

        self.assertTrue(os.path.isfile(fp))

        new_life = Island(filepath=fp)

        self.assertDictEqual(new_life.function_params, life.function_params)
        self.assertListEqual(new_life.population, life.population)
        self.assertEqual(new_life.name, life.name)

        new_life.evolve(starting_generation=new_life.generation_count)

        self.assertEqual(new_life.generation_count, life.generation_count*2)
        os.remove(fp)


class TestException(unittest.TestCase):

    def test_raise_exception(self):
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])
        ind = Individual(fitness_raise_error, name="Adam", chromosome=gen)

        life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, verbose=True)

        try:
            life.initialise(ind, population_size=5)
        except:
            self.assertTrue(True)

class TestCheckpoints(unittest.TestCase):

    def test_save_load(self):
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])
        ind = Individual(fitness, name="Adam", chromosome=gen)

        life = Island({'x': 0.6, 'y': 0.2}, save_checkpoint_level=2)

        life.initialise(ind, population_size=5)

        life.evolve()



import copy
from typing import Callable

def initialise_population_random(adam, n : int = 10, chromosome_create_func : Callable = None, island=None):
    """
    Classic random initialisation function to create a pool of `n` individuals from a starting Individual `adam`.

    Args:
        adam (Individual): An individual already initialised with a chromosome.
        n (int): Population size (default = 10).
        chromosome_create_func (): A function to create the chromosome of new individuals. If none, simply deep copy adam's. (default = None).
        island (Island): Needed to wrap to `create_chromosome` and `create_individual` methods (default = None).

    Returns:
        list: Population members.
    """
    population = list()

    for i in range(n - 1):
        if not chromosome_create_func is None:
            chromosome = chromosome_create_func()
        else:
            chromosome = island.create_chromosome(genes=[x.randomise_new() for x in adam.chromosome],
                                              gene_verify_func=adam.chromosome.gene_verify_func,
                                              chromosome_properties=copy.deepcopy(adam.chromosome.get_properties()))
        eve = island.create_individual(adam.fitness_function,
                                       chromosome=chromosome,
                                       species_type=adam.species_type,
                                       individual_properties=copy.deepcopy(adam.get_properties()))
        population.append(eve)

    population.append(adam)

    return population

def initialise_population_mutation_function(adam, n : int = 10, chromosome_create_func : Callable = None, mutation_params : dict = None, island=None):
    """
    Random initialisation function to create a pool of `n` individuals from a starting Individual `adam`, but uses the ``island.mutation function`` to perform the randomisation.

    Args:
        adam (Individual): An individual already initialised with a chromosome.
        n (int): Population size (default = 10).
        chromosome_create_func (): A function to create the chromosome of new individuals. If none, simply deep copy adam's. (default = None).
        mutation_params (dict): The params of the mutation function, these usually need to be higher values, such as higher ``prob`` value.
        island (Island): Needed to wrap to `create_chromosome` and `create_individual` methods (default = None).

    Returns:
        list: Population members.
    """

    population = list()

    if mutation_params:
        _mutation_params = mutation_params
    else:
        _mutation_params = {}

    for i in range(n - 1):
        if not chromosome_create_func is None:
            chromosome = chromosome_create_func()
        else:
            chromosome = copy.deepcopy(adam.chromosome)
        eve = island.create_individual(adam.fitness_function,
                                       chromosome=chromosome,
                                       species_type=adam.species_type,
                                       individual_properties=copy.deepcopy(adam.get_properties()))
        eve = island.mutation(island=island, individual=eve, **_mutation_params)
        population.append(eve)

    population.append(adam)

    return population


import copy
import os
from datetime import datetime

def clone_classic(individuals : list, island=None):
    """
    Classic cloning function, making a deep copy of population list.

    Args:
        individuals (list): Population members to copy.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        list: Carbon copy of population.
    """
    return copy.deepcopy(individuals)

def default_save_checkpoint_function(event, island):
    """
    Default checkpoint saving function. Used to save the Island periodically after certain events.

    Args:
        event (str): The evolutionary event (crossover, eval etc).
        island (Island): The island to save.

    """
    if not os.path.isdir(island.checkpoints_dir):
        os.mkdir(island.checkpoints_dir)
    if not os.path.isdir(f'{island.checkpoints_dir}/{island.name}'):
        os.mkdir(f'{island.checkpoints_dir}/{island.name}')
    fp = f'{island.checkpoints_dir}/{island.name}/{datetime.utcnow().strftime("%H-%M-%S")}_{event}_checkpoint.pkl'
    island.verbose_logging(f'checkpoint: file {fp}')
    island.save_island(fp)

def post_evolution_function_save_all(island):
    """
    Simple function to save (pickle) every population member after completing evolution.

    Args:
        island (Island): Island to get ``population`` from.
    """
    for p in island.population:
        p.save_individual(filepath=f'individual_{p.name}.pkl')

class GeneticAlgorithmError(Exception):
    """
    Encapsulating graceful exception handling during evolutionary runs.

    Args:
        message (str): Message to print.
        exit (bool): Whether to hard exit the process or not (default = False).
    """

    def __init__(self, message : str, exit : bool = False):
        self.message = message
        if exit:
            print(f"GeneticAlgorithmError: {self.message}")
            quit(1)


    def __str__(self):
        return f"GeneticAlgorithmError: {self.message}"